from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import hex
from builtins import super
from future import standard_library
standard_library.install_aliases()

import uuid

from valvepcf.constants import OPERATORS


class PcfNode(object):
    """
    An abstract class to represent different types of pcf elements
    and hold their collections of attributes
    """

    def __init__(self, ename, etype, edesc='', euuid=None):
        """
        :param ename: the element's name
        :type ename: str
        :param etype: the element's type
        :type etype: str
        :param edesc: the element's description
        :type edesc: str, optional
        :param euuid: the element's uuid (autogenerated if not provided)
        :type euuid: str, optional
        """

        self._type = etype  #: :type: (str) - The element's type.
        self._name = ename or 'untitled'  #: :type: (str) - The element's name.
        #: :type: (str) - The element's uuid (generally autogenerated).
        self._uuid = euuid or uuid.uuid4().bytes
        #: :type: (str) - The element's description.
        self._desc = edesc or 'DmElement'
        #: :type: (list[PcfAttribute]) - The element's list of attributes.
        self.attributes = []

    def __repr__(self, indent=0, operator=''):
        """A printable summary of the PcfNode and its child nodes.

        :returns: A Python formated string.
        :rtype: str
        """

        module_ = type(self).__module__
        name_ = type(self).__name__
        type_ = operator or self._type
        ret = "  " * indent
        ret += "<{0}.{1} {2}".format(module_, name_, type_) + \
               " named '{0}' at {1}>".format(self._name, hex(id(self)))
        return ret


class PcfRootNode(PcfNode):
    """
    PcfRootNode is the top level structure which includes a collection of
    :py:class:`PcfSystemNode<PcfSystemNode>` as well as pcf format and schema versions.
    """

    def __init__(self, ename, etype, edesc='', euuid=None):

        super().__init__(ename, etype, edesc, euuid)

        self.binary_format = 'dmx'  #: :type: (str) Always 'dmx'
        self.binary_version = 3  #: :type: (int) Ranges from 2 to 5
        self.pcf_format = 'pcf'  #: :type: (str) 'dmx' or 'pcf'
        self.pcf_version = 2  #: :type: (int) Ranges from 1 to 2

        #: :type: (list[PcfSystemNode]) List of particle systems
        self.systems = []

        self._unaccounted_strings = []
        self._order = None

    def __repr__(self, indent=0):
        ret = super().__repr__()
        for syst in self.systems:
            ret += '\n' + syst.__repr__(indent + 1)
        return ret


class PcfSystemNode(PcfNode):
    """PcfSystemNode is used to define a pcf system which regroups multiple
    :py:class:`PcfOperatorNode<PcfOperatorNode>`. Operators are separated in specialized lists."""

    def __init__(self, ename, etype, edesc='', euuid=None):
        super().__init__(ename, etype, edesc, euuid)

        #: :type: (list[PcfOperatorNode]) - list of renderers
        self.renderers = []
        #: :type: (list[PcfOperatorNode]) - list of operators
        self.operators = []
        #: :type: (list[PcfOperatorNode]) - list of initializers
        self.initializers = []
        #: :type: (list[PcfOperatorNode]) - list of emmiters
        self.emitters = []
        #: :type: (list[PcfRefNode]) - list of childrens
        self.children = []
        #: :type: (list[PcfOperatorNode]) - list of forces
        self.forces = []
        #: :type: (list[PcfOperatorNode]) - list of constraints
        self.constraints = []

    def __repr__(self, indent=0):
        ret = super().__repr__(indent)
        for c in reversed(OPERATORS):
            for node in getattr(self, c):
                ret += '\n' + node.__repr__(indent + 1, c)
        return ret


class PcfOperatorNode(PcfNode):
    """PcfOperatorNode is used to hold specific collection of attributes for
    :py:class:`PcfSystemNode<PcfSystemNode>`."""


class PcfRefNode(PcfNode):
    """PcfRefNode is used to create child relationships between particle systems."""

    def __init__(self, ename, etype, edesc='', euuid=None):
        super().__init__(ename, etype, edesc, euuid)

        #: :type: (PcfSystemNode) - reference an other system
        self.ref = None


class PcfAttribute(object):
    """
    PcfAttribute represents an attribute held by an element.
    """

    def __init__(self, name, type, data):
        """
        :param name: The attribute's name.
        :param type: The attribute's type.
        :param data: The attribute's value.
        """
        #: :type: (str) - Name of the attribute
        self._name = name
        #: :type: (int) - Ranges (1-28), defines the type of :py:attr:`_data<_data>`
        self._type = type
        #: The attribute's value, can be one of many different types, must match :py:attr:`_type<_type>`
        self._data = data



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import range
from builtins import open
from builtins import next
from future import standard_library
standard_library.install_aliases()

from valvepcf.structs import *
from valvepcf.classes import *
from valvepcf.constants import OPERATORS


def save_pcf(pcf, dest):
    with open(dest, 'wb') as f:
        data = unload_pcf(pcf)
        PCF.build_stream(data, f)


def unload_pcf(pcf):
    output = {'versions': (pcf.binary_version,
                           pcf.pcf_format,
                           pcf.pcf_version),
              'strings': [],
              'elements': [],
              'attributes': []}

    relisted = []
    relisted.append(pcf)
    systems = list(pcf.systems)

    def register_elem(elem, order, elemlist=[]):

        elemlist.append(elem)

        if isinstance(elem, PcfSystemNode):
            for o in reversed([o for o in order if o in OPERATORS]):
                if o == 'children':
                    for child in elem.children:
                        elemlist = register_elem(child, order, elemlist)
                else:
                    items = getattr(elem, o, None)
                    if items:
                        elemlist = elemlist + items

        elif isinstance(elem, PcfRefNode):
            if elem.ref in systems:
                systems.remove(elem.ref)
            if elem.ref and elem.ref not in elemlist:
                elemlist = register_elem(elem.ref, order, elemlist)

        return elemlist

    while(systems):
        system = systems.pop(0)
        relisted = register_elem(system, pcf._order, relisted)

    def pcfnode_to_elem(pcfnode):
        return {'elementType': pcfnode._type,
                'elementDesc': pcfnode._desc,
                'elementName': pcfnode._name,
                'elementUUID': pcfnode._uuid}

    def new_attr(attr_name, attr_type, attr_data):
        return {'attributeName': attr_name,
                'attributeType': attr_type,
                'attributeData': attr_data}

    def pcfattr_to_attr(pcfattr):
        return new_attr(pcfattr._name, pcfattr._type, pcfattr._data)

    def ref_to_attr(node, attr_name):
        return new_attr(attr_name, 1, relisted.index(node))

    def refs_to_attr(node_list, attr_name):
        ids = [relisted.index(node) for node in node_list]
        return new_attr(attr_name, 15, ids)

    def node_attrs(node):
        nattrs = []

        list_attributes = [x for x in node.attributes if x._type >= 15]
        norm_attributes = [x for x in node.attributes if x._type < 15 and
                           x._type != 1]

        def register_attributes():
            for nattr in list_attributes:
                nattrs.append(pcfattr_to_attr(nattr))

            for nattr in norm_attributes:
                nattrs.append(pcfattr_to_attr(nattr))

        def register_root():
            attrs_types = list(pcf._order)

            for a_t in attrs_types:
                if a_t == 'attributes':
                    register_attributes()
                elif a_t == 'particleSystemDefinitions':
                    nattrs.append(refs_to_attr(node.systems,
                                               'particleSystemDefinitions'))

        def register_children():
            attrs_types = list(pcf._order)
            for a_t in attrs_types:
                if a_t == 'attributes':
                    register_attributes()
                elif a_t == 'children' and node.ref:
                    nattrs.append(ref_to_attr(node.ref, 'child'))

        def register_system_node():
            attrs_types = list(pcf._order)

            if 'material' in attrs_types:
                material_attr = next(
                    (x for x in norm_attributes if x._name == 'material'), None)
                if material_attr:
                    norm_attributes.remove(material_attr)
            if 'color' in attrs_types:
                color_attr = next(
                    (x for x in norm_attributes if x._name == 'color'), None)
                if color_attr:
                    norm_attributes.remove(color_attr)

            for a_t in attrs_types:
                # we golfin bois!
                if a_t == 'material' and material_attr:
                    nattrs.append(pcfattr_to_attr(material_attr))
                if a_t == 'color' and color_attr:
                    nattrs.append(pcfattr_to_attr(color_attr))
                if a_t == 'attributes':
                    register_attributes()
                else:
                    d = getattr(node, a_t, None)
                    if d != None and a_t in pcf._data.strings:
                        nattrs.append(refs_to_attr(d, a_t))

        if isinstance(node, PcfSystemNode):
            register_system_node()
        elif isinstance(node, PcfRootNode):
            register_root()
        elif isinstance(node, PcfRefNode):
            register_children()
        else:
            register_attributes()

        return nattrs

    def register_string(pcf_string, minver=0):
        # Register a string
        if isinstance(pcf_string, int):
            return pcf_string
        if pcf.binary_version >= minver:
            if pcf_string not in output['strings']:
                output['strings'].append(pcf_string)
                if pcf_string in ['functionName', 'child'] and \
                   'type_dictionary' in pcf._unaccounted_strings:
                    # dont' know what this string is about, but it's always there
                    register_string('type_dictionary')
            return output['strings'].index(pcf_string)
        else:
            return pcf_string

    def register_element_strings(elem):
        # Register strings of an element
        elem['elementType'] = register_string(elem['elementType'])
        elem['elementDesc'] = register_string(elem['elementDesc'], 5)
        elem['elementName'] = register_string(elem['elementName'], 4)

    def register_attribute_strings(attr):
        # Register strings of an attribute
        attr['attributeName'] = register_string(attr['attributeName'])
        if (attr['attributeType'] == 5):
            attr['attributeData'] = register_string(attr['attributeData'], 4)

    for nodeid in range(len(relisted)):

        # make element
        n_elem = pcfnode_to_elem(relisted[nodeid])
        output['elements'].append(n_elem)
        register_element_strings(n_elem)

        # make attributes
        n_attrs = node_attrs(relisted[nodeid])
        output['attributes'].append(n_attrs)
        for attr in reversed(n_attrs):
            register_attribute_strings(attr)

        # additional garbo
        if nodeid == 0 and 'name' in pcf._unaccounted_strings:
            register_string('name')

    if 'type_dictionary' in pcf._unaccounted_strings and pcf.binary_version < 4:
        register_string('type_dictionary')

    return output



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import range
from builtins import next
from builtins import open
from future import standard_library
standard_library.install_aliases()

from valvepcf.structs import *
from valvepcf.classes import *
from valvepcf.constants import DEFAULT_ORDER, OPERATORS

import collections
try:
    collectionsAbc = collections.abc
except AttributeError:
    collectionsAbc = collections


def load_pcf(pcf):

    with open(pcf.source_path, 'rb') as f:
        data = PCF.parse_stream(f)

    nodes = []
    attrs = []

    tracked_strings = TrackedStrings(data.strings)

    # load elements and attributes
    for i in range(len(data.elements)):
        attrs.append(load_attributes(data.attributes[i], tracked_strings))
        nodes.append(load_element(data.elements[i], tracked_strings))

    root = nodes[0]
    load_versions(root, data)

    # make references between nodes, load their attributes
    for i in range(len(nodes)):
        if nodes[i]._type in ['DmeElement', 'DmElement']:
            load_node_systems(nodes[i], attrs[0], nodes)
        elif nodes[i]._type == 'DmeParticleChild':
            load_node_ref(nodes[i], attrs[i], nodes)
        elif nodes[i]._type == 'DmeParticleSystemDefinition':
            load_node_operators(nodes[i], attrs[i], nodes)
        load_node_attributes(nodes[i], attrs[i])

    root._unaccounted_strings = list(tracked_strings.unused)
    root._order = load_order(data, tracked_strings)

    pcf._data = data
    load_root(pcf, root)


class TrackedStrings(collectionsAbc.MutableMapping, list):
    # provides the list of parsed strings while tracking their usage
    def __init__(self, strings):
        self._strs = list(strings)
        self.unused = list(strings)

    def __getitem__(self, sid):
        if isinstance(sid, int):
            self.unused = [x for x in self.unused if x is not self._strs[sid]]
            return self._strs[sid]
        else:
            return sid


def ensure_once(itemlist, item):
    if item not in itemlist:
        itemlist.append(item)


def load_versions(node, data):
    node.binary_format = 'dmx'
    node.binary_version = data.versions[0]
    node.pcf_format = data.versions[1]
    node.pcf_version = data.versions[2]


def load_order(data, strings):
    ORDERABLE_ATTRS = ['material', 'color']
    root_order = []

    for root_attr in data.attributes[0]:
        referential = root_attr.attributeType == 15
        name = 'attributes' if not referential else 'particleSystemDefinitions'
        ensure_once(root_order, name)
    for name in [o for o in DEFAULT_ORDER if o not in OPERATORS]:
        ensure_once(root_order, name)

    element = next((x for x in data.elements if strings[x.elementType] ==
                    'DmeParticleSystemDefinition'), None)
    if not element:
        return DEFAULT_ORDER

    elem_attrs = data.attributes[data.elements.index(element)]

    order = []
    for attr in elem_attrs:
        addname = strings[attr.attributeName]
        if addname not in ORDERABLE_ATTRS and attr.attributeType not in [1, 15]:
            addname = 'attributes'

        # remove ORDERABLE_ATTRS if they are contiguous with 'attributes'
        last = order[-1] if order else None
        if last in ORDERABLE_ATTRS and addname == 'attributes':
            order = [s for s in order if s not in ORDERABLE_ATTRS]
        # skip ORDERABLE_ATTRS if they are contiguous with 'attributes'
        elif last == 'attributes' and addname in ORDERABLE_ATTRS:
            continue

        if addname == 'attributes':
            for name in root_order:
                ensure_once(order, name)
            continue

        ensure_once(order, addname)

    return order or DEFAULT_ORDER


def load_node_ref(node, nattrs, nodes):
    node.ref = next((nodes[attr._data] for attr in nattrs
                     if attr._type == 1), None)


def load_node_systems(node, nattrs, nodes):
    for attr in nattrs:
        if attr._type == 15:
            for eid in attr._data:
                node.systems.append(nodes[eid])


def load_node_operators(node, nattrs, nodes):
    for attr in [x for x in nattrs if x._type == 15 and x._name in OPERATORS]:
        for eid in attr._data:
            getattr(node, attr._name, None).append(nodes[eid])


def load_node_attributes(node, nattrs):
    node.attributes = [attr for attr in nattrs if attr._type not in [1, 15]]


TYPEMAP = {
    'DmElement': PcfRootNode,
    'DmeElement': PcfRootNode,
    'DmeParticleSystemDefinition': PcfSystemNode,
    'DmeParticleOperator': PcfOperatorNode,
    'DmeParticleChild': PcfRefNode,
}


def load_element(elem, strings):
    ename = strings[elem.elementName]
    etype = strings[elem.elementType]
    edesc = strings[elem.elementDesc]
    euuid = elem.elementUUID
    PClass = TYPEMAP[etype] if etype in TYPEMAP.keys() else PcfNode
    return PClass(ename, etype, edesc, euuid)


def load_attribute(attr, strings):
    aname = strings[attr.attributeName]
    atype = attr.attributeType
    adata = attr.attributeData
    if atype == 5:
        adata = strings[attr.attributeData]
    elif (atype >= 8 and atype <= 14) or \
            atype == 15 or atype > 22:
        adata = list(attr.attributeData)
    return PcfAttribute(aname, atype, adata)


def load_attributes(attrs, strings):
    return [load_attribute(a, strings) for a in attrs]


def load_root(pcf, root):
    pcf._name = root._name
    pcf._type = root._type
    pcf._uuid = root._uuid
    pcf._desc = root._desc

    pcf.binary_format = root.binary_format
    pcf.binary_version = root.binary_version
    pcf.pcf_format = root.pcf_format
    pcf.pcf_version = root.pcf_version

    pcf.systems = root.systems
    pcf.attributes = root.attributes

    pcf._unaccounted_strings = root._unaccounted_strings
    pcf._order = root._order



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
from valvepcf.pcf import *  # NOQA: E402



from distutils.core import setup
setup(
    name='valvepcf',
    packages=['valvepcf'],
    version='1.1.0',
    license='gpl-3.0',
    description='A library to parse .pcf files used by the source engine.',
    author='Maxime Dupuis',
    author_email='mdupuis@hotmail.ca',
    url='https://github.com/pySourceSDK/ValvePCF',
    download_url='https://github.com/pySourceSDK/ValvePCF/archive/v1.1.0.tar.gz',
    keywords=['pcf', 'source', 'sourcesdk', 'hammer', 'valve'],
    install_requires=['construct', 'future'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
)
